/*
 *	Generic svgalib moving objects.
 * Note that .bitmap is just a pointer to bitmap malloced elsewhere.
 * First create [&recolor] a bitmap, then create a bob using it.
 */

#include <stdlib.h>
#include <stdio.h>
#include <vga.h>
#include <vgagl.h>
#include <unistd.h>

#include "hr_defines.h"
#include "hr_bitmaps.h"
#include "hr_general.h"
#include "hr_mouse.h"


struct hr_bob{
	int flags;
	int x;
	int y;
	int hot_x;
	int hot_y;
	struct hr_bitmap *bitmap;
	struct hr_bitmap *tempmap;
	struct hr_bob *next;
	struct hr_bob *prev;
};


/* Deletes a bob. Tempmap is freed, bitmap data is not touched.
 * Mouse pointer cannot be deleted using this func.*/
struct hr_bob *hr_delete_bob(struct hr_bob *b)
{
   if((b!=NULL)&&(b!=hr_mouse.bob)) {
		if(!(b->flags & BOB_NO_TEMPMAP))
		  hr_delete_bitmap(b->tempmap);
		/* Set pointers .*/
		if(b->prev!=NULL)
		  b->prev->next=(b->next);
		if(b->next!=NULL)
		  b->next->prev=(b->prev);
		free(b);
		return b;
	}
	return NULL;
}


/* Creates a bob and returns pointer to it. */
struct hr_bob *hr_create_bob(struct hr_bitmap *bitmap,int hot_x,int hot_y,
									  int flags)
{
	struct hr_bob *b=NULL;
	unsigned char *t;
	if(flags & GFX_COLOR){
		b=malloc(sizeof(struct hr_bob));
		/* If any bobs already exist or
		 * mouse pointer is NOT required to stay on top of other bobs all
		 * the time, new bob is added as first items in list. */
		if((hr.firstbob==NULL)||(!(hr_mouse.flags & MOUSE_STAYS_ON_TOP))){
			if(hr.firstbob==NULL)
			  b->next=NULL;
			else
			  b->next=hr.firstbob;
			hr.firstbob=b;
			b->prev=NULL;
		}
		/* If bobs already exist and mouse pointer must stay on top,
		 * add new bob as 2nd item in list.*/
		else{
			hr.firstbob->next->prev=b;
			b->next=hr.firstbob->next;
			hr.firstbob->next=b;
			b->prev=hr.firstbob;
		}
		/* Set pointers and variables.*/
		b->flags=flags;
		b->x=0;
		b->y=0;
		b->hot_x=hot_x;
		b->hot_y=hot_y;
		b->bitmap=bitmap;
		/* This is needed to save the background gfx.*/
		t=malloc((bitmap->width)*(bitmap->height));
		if(flags & BOB_NO_TEMPMAP)
		  b->tempmap=NULL;
		else
		  b->tempmap=hr_create_bitmap(t,NULL,bitmap->width,bitmap->height,
												256,NONE);
	}
	return b;
}


/* Updates a bob. */
void hr_update_bob(struct hr_bob *b,int x,int y,
						 struct hr_bitmap *bitmap,int mode)
{
	int ox,oy;
	/**/
	if(b!=NULL) {
		ox=b->x;
		oy=b->y;
		/* Do not change bob's position if it is inited/removed ONLY */
		switch(mode){
		 case BOB_INITONLY:
		 case BOB_REMOVEONLY:
			x=b->x;
			y=b->y;
			break;
		 default:
			b->x=x;
			b->y=y;
			break;
		}
		/* If the bob is wanted to be animated...*/
		if(bitmap!=NULL)
		  b->bitmap=bitmap;
		/* case BOB_NOIMG is ignored.. */
		switch(mode) {
		 case BOB_INIT:
		 case BOB_INITONLY:
			/* Initalize bob after some gfx is changed.*/
			if(!(b->flags & BOB_INITIALIZED)) {
				b->flags |= BOB_INITIALIZED;
				if(!(b->flags & BOB_NO_TEMPMAP))
				  gl_getbox(x-(b->hot_x),y-(b->hot_y),
								b->tempmap->width,b->tempmap->height,
								b->tempmap->data);
				gl_putboxmask(x-(b->hot_x),y-(b->hot_y),
								  b->bitmap->width,b->bitmap->height,b->bitmap->data);
				break;
			}
		 case BOB_SMART:
			/* Normal mode. Bob moved on top of unchanging background. */
			/* Now save the pointer background and paste pointer to screen.*/
			if((b->flags & BOB_INITIALIZED) &&
				((ox!=x)||(oy!=y))) {
				/* Normal mode. Bob moved on top of unchanging background.
				 * First restores old background, then saves new background
				 * and draws Bob pointer. */
				if(!(b->flags & BOB_NO_TEMPMAP)){
					gl_putbox(ox-(b->hot_x),oy-(b->hot_y),
								 b->tempmap->width,b->tempmap->height,
								 b->tempmap->data);
					gl_getbox(x-(b->hot_x),y-(b->hot_y),
								 b->tempmap->width,b->tempmap->height,
								 b->tempmap->data);
				}
				gl_putboxmask(x-(b->hot_x),y-(b->hot_y),
								  b->bitmap->width,b->bitmap->height,
								  b->bitmap->data);
			}
			break;
		 case BOB_DUMB:
			/* This is used when the whole screen is constantly redrawn. */
			gl_putboxmask(x-(b->hot_x),y-(b->hot_y),
							  b->bitmap->width,b->bitmap->height,
							  b->bitmap->data);
			break;
		 case BOB_REMOVE:
		 case BOB_REMOVEONLY:
			/* Remove Bob from screen */
			if(b->flags & BOB_INITIALIZED) {
				b->flags &= ~(BOB_INITIALIZED);
				if(!(b->flags & BOB_NO_TEMPMAP))
				  gl_putbox(ox-(b->hot_x),oy-(b->hot_y),
								b->tempmap->width,b->tempmap->height,
								b->tempmap->data);
			}
			break;
		}
	}
}


/* Checks bobs from "first" onwards. Pointer to first one that overlaps with
 * bob "what" is returned. */
struct hr_bob *hr_bob_overlap(struct hr_bob *what,struct hr_bob *first)
{
	struct hr_bob *bob=first,*lapbob=NULL;
	int hh,ww;
	int wx1,wx2,wy1,wy2;
	int bx1,bx2,by1,by2;
	if(what!=NULL){
		ww=what->bitmap->width;
		hh=what->bitmap->height;
		/* Hot spots are taken into account.*/
		wx1=(what->x)-(what->hot_x);
		wx2=(what->x)+(what->bitmap->width)-1-(what->hot_x);
		wy1=(what->y)-(what->hot_y);
		wy2=(what->y)+(what->bitmap->height)-1-(what->hot_y);
		while(bob!=NULL){
			if(bob!=what){
				bx1=(bob->x)-(bob->hot_x);
				bx2=bx1+(bob->bitmap->width)-1;
				/* Overlap horizontally?*/
				if(((wx1<=bx2)||(wx2<=bx2))&&
					((wx1>=bx1)||(wx2>=bx1)))
				  {
					  by1=(bob->y)-(bob->hot_y);
					  by2=by1+(bob->bitmap->height)-1;
					  /* Overlap vertically?*/
					  if(((wy1<=by2)||(wy2<=by2))&&
						  ((wy1>=by1)||(wy2>=by1)))
						 {
							 lapbob=bob;
							 break;
						 }
				  }
			}
			bob=bob->next;
		}
	}
	return lapbob;
}


/* Checks bobs from "first" onwards. Pointer to first one that overlaps with
 * bob "what" is returned. */
struct hr_bob *hr_bob_collide(struct hr_bob *what,struct hr_bob *first)
{
	struct hr_bob *bob;
	int x1,y1,x2,y2,cx1,cy1,cx2,cy2,ch,cw,ox1,oy1,ox2,oy2;
	int xx,yy;
	unsigned char *p1,*p2,*sp1,*sp2;
	/* Get first bob that overlaps with "what", then,
	 * if such bob was found, See, using masks, whether the bobs overlap.*/
	if((bob=hr_bob_overlap(what,first))){
		/* Calc topleft coordinates of both bobs */
		x1=(what->x)-(what->hot_x);
		y1=(what->y)-(what->hot_y);
		x2=(bob->x)-(bob->hot_x);
		y2=(bob->y)-(bob->hot_y);
		/* Calc coordinates of area to be checked.*/
		cx1=MAX(x1,x2);
		cx2=MIN(x1+(what->bitmap->width)-1,x2+(bob->bitmap->width)-1);
		cy1=MAX(y1,y2);
		cy2=MIN(y1+(what->bitmap->height)-1,y2+(bob->bitmap->height)-1);
		/* Calc width and height of area to be checked */
		cw=cx2-cx1+1;
		ch=cy2-cy1+1;
		/* Calc offsets of area to be checked within both bobs.*/
		ox1=cx1-x1;
		ox2=cx2-x2;
		oy1=cy1-y1;
		oy2=cy2-y2;
		/* Do logical AND to overlapping bytes. If the result is non-zero,
		 * we know for sure that the bobs collide.*/
		/* Calc starting addresses. */
		sp1=(what->bitmap->data)+oy1*(what->bitmap->width)+ox1;
		sp2=(bob->bitmap->data)+oy2*(bob->bitmap->width)+ox2;
		for(yy=0;yy<ch;yy++){
			p1=sp1;
			p2=sp2;
			for(xx=0;xx<cw;xx++){
				if(*p1 & *p2)
				  return bob;
				++p1;
				++p2;
			}
			/* Starting addresses move to next line.*/
			sp1+=(what->bitmap->width);
			sp2+=(bob->bitmap->width);
		}
		bob=bob->next;
	}
	return NULL;
}


/* Moves bob to front */
void bob_to_front(struct hr_bob *bob)
{
	if((bob!=NULL)&&
		(bob!=hr.firstbob))
	  {
		  /* Make next/prev pointers in list "jump over" bob */
		  if((bob->prev)!=NULL)
			 bob->prev->next=(bob->next);
		  if((bob->next)!=NULL)
			 bob->next->prev=(bob->prev);
		  /* If mouse pointer is forced to always be the frontmost bob,
			* *bob is moved as 2nd item in the list. */
		  if(hr_mouse.flags & MOUSE_STAYS_ON_TOP){
			  bob->next=(hr.firstbob->next);
			  bob->prev=hr.firstbob;
			  hr.firstbob->next=bob;
		  }
		  else{
			  bob->next=hr.firstbob;
			  bob->prev=NULL;
			  hr.firstbob=bob;
		  }
		  
	  }
}


/* Moves bob to back */
void bob_to_back(struct hr_bob *bob)
{
	struct hr_bob *tb;
	if(bob!=NULL){
		/* Frontmost cannot be moved to back if frontmost bob is the
		 * mouse pointer.*/
		if(!((hr_mouse.flags & MOUSE_STAYS_ON_TOP)&&
			  (bob==hr.firstbob)))
		  {
			  /* Make next/prev pointers in list "jump over" bob */
			  if((bob->prev)!=NULL)
				 bob->prev->next=(bob->next);
			  if((bob->next)!=NULL)
				 bob->next->prev=(bob->prev);
			  /**/
			  tb=hr.firstbob;
			  while(tb->next!=NULL)
				 tb=tb->next;
			  tb->next=bob;
			  bob->next=NULL;
			  bob->prev=tb;
		  }
	}
}


/* Moves bob under "topbob" */
void bob_to_after(struct hr_bob *bob,struct hr_bob *topbob)
{
	struct hr_bob *tb;
	if((bob!=NULL)&&
		(topbob!=NULL)&&
		(bob!=topbob)){
		/* Frontmost cannot be moved to back if frontmost bob is the
		 * mouse pointer.*/
		if(!((hr_mouse.flags & MOUSE_STAYS_ON_TOP)&&
			  (bob==hr.firstbob)))
		  {
			  /* Make next/prev pointers in list "jump over" bob */
			  if((bob->prev)!=NULL)
				 bob->prev->next=(bob->next);
			  if((bob->next)!=NULL)
				 bob->next->prev=(bob->prev);
			  /**/
			  tb=hr.firstbob;
			  while(tb!=NULL)
				 tb=tb->next;
			  bob->next=tb->next;
			  if((tb->next)!=NULL)
				 tb->next->prev=bob;
			  tb->next=bob;
			  bob->prev=tb;
		  }
	}
}
